# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from collections import OrderedDict
from cStringIO import StringIO
from grp import getgrnam
from network_docopt import sort_for_humans
from pprint import pformat
from pwd import getpwnam
from socket import gethostbyaddr
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate  # This is an add-on from Pip.
import errno
import filecmp                 # Using filecmp.cmp.  Note that cmp() is a built-in function.
import json
import logging
import logging.handlers
import os
import re
import shutil
import sys

log = logging.getLogger('netd')

PENDING_COMMANDS_FILE = '/var/run/nclu/pending_commands.txt'
NCLU_VERSION = 1.0

ifupdown2_glob_regexs = [re.compile(r"([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\]([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\](.*)"),
                         re.compile(r"([A-Za-z0-9\-]+[A-Za-z])(\d+)\-(\d+)(.*)"),
                         re.compile(r"([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\](.*)")]


# Each show_pending needs to return the correct state
# since we could have one wrapper with an error.
# In that case we should not commit any changes at all.
PENDING_COMMIT_NOCHANGE = 0
PENDING_COMMIT_CHANGE   = 1
PENDING_COMMIT_ERROR    = 2

class Capturing(list):

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout


def files_match(fileA, fileB):
    # TODO - Depending on how and where this is used downstream, I might be able to replace this
    # function with filecmp.cmp, which I'm calling here.
    if not os.path.isfile(fileA) and not os.path.isfile(fileB):
        return True   # Neither file exists.  TODO - Should this raise an exception instead?

    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        # TODO - Should this raise an exception instead?
        return False  # Only one file exists, so they cannot be equal.

    return filecmp.cmp(fileA, fileB)


def print_file(filename):
    with open(filename, 'r') as f:
        print f.read()


def print_diff(fileA, fileB, color_diffs):
    """
    Print the diff between fileA vs. fileB with colored output, and
    return True if we printed anything.
    """
    # TODO - Can I delete code by using the difflib standard library package?  This would remove
    # the subprocess call, but difflib.Differ().compare(...) has a different return value syntax
    # than diff.  The former does not include lines beginning with --- or +++.

    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        return False  # One or both files do not exist.

    try:
        check_output(['/usr/bin/diff',
                      '-U', '10',
                      '--ignore-blank-lines',
                      '--ignore-space-change',
                      fileA,
                      fileB])
        return False
    except CalledProcessError as e:

        # Print the deletes in red and the adds in green
        for line in e.output.splitlines():
            if color_diffs:
                if line.startswith('---') or line.startswith('+++'):
                    print "\033[94m%s\033[0m" % line
                elif line.startswith('-'):
                    print "\033[91m%s\033[0m" % line
                elif line.startswith('+'):
                    print "\033[92m%s\033[0m" % line
                else:
                    print line
            else:
                print line
        return True


def chown_chgroup(filename, user, group):
    quagga_uid = getpwnam(user).pw_uid
    quagga_gid = getgrnam(group).gr_gid
    os.chown(filename, quagga_uid, quagga_gid)


def file_contains(filename, target_line):
    """
    Return True if the given file contains the given line.  Otherwise, return False.
    """
    if os.path.isfile(filename):
        with open(filename) as f:
            line = f.readline()
            while line:
                if line.strip() == target_line.strip():
                    return True
                line = f.readline()
    return False


def ifname_is_valid(ifname):
    """
    Enforce the rules for interface names defined in the kernel's dev_valid_name().
    Valid interface names can have up to 15 visible characters plus a null terminator.
        http://stackoverflow.com/questions/24932172/what-length-can-a-network-interface-name-have#29398765
    """
    IF_NAMESIZE = 15
    ILLEGAL_CHARACTERS = ('/', ' ', '%', '*', '?', '^', '$', '+', '[', ']', '{', '}', '(', ')', '\\')
    ILLEGAL_NAMES = ('.', ':')  # Illegal names comprised of legal characters

    ifnames = ifname if isinstance(ifname, list) else [ifname]

    for ifacename in ifnames:
        if ifacename in ILLEGAL_NAMES:
            print "'{0}' is an invalid interface name (cannot use '.' or ':').".format(ifacename)
            return False

        for character in ILLEGAL_CHARACTERS:
            if character in ifacename:
                print "'{0}' is an invalid interface name (cannot use '{1}').".format(ifacename, "', '".join(ILLEGAL_CHARACTERS))
                return False

        if len(ifacename) > IF_NAMESIZE:
            print "'{0}' is an invalid interface name (more than {1} characters).".format(ifacename, IF_NAMESIZE)
            return False

    return True


def ifname_is_globish(ifname):
    """
    This assumes that ifname is not a glob, as verified by calling ifname_is_glob().

    Return True if the ifname contains characters that one might use when
    attempting to enter a glob.
    """
    GLOBISH_CHARACTERS = (',', '[', ']', '(', ')', '%', '*')

    glob_msg = """
'{0}' looks sort of like a glob but does not use valid glob syntax.
Globs use '-' to specify a range of interfaces.

Examples:
    swp1-4
    swp1-4,swp8,swp9
    swp1-4,8,9
    swp1-4,swp8-9
    swp1,swp8,swp9
    swp1,8,9
    swp1,swp8-9

""".format(ifname)

    if re.match(r"""^\w+0\d+(?:-0\d+)?$""", ifname) is not None:
        return False

    if re.match(r"""^swp\d+-\w+\d+$""", ifname) is not None:
        print glob_msg
        return True

    for character in GLOBISH_CHARACTERS:
        if character in ifname:
            print glob_msg
            return True

    return False


def ifupdown2_ifname_is_glob(ifname):
    """
    From ifupdown2's ifupdownaddons/modulebase.py
    """
    if (ifupdown2_glob_regexs[0].match(ifname) or
        ifupdown2_glob_regexs[1].match(ifname) or
        ifupdown2_glob_regexs[2].match(ifname)):
        return True
    return False


def ifupdown2_parse_glob(ifacename, expr):
    """
    From ifupdown2's ifupdownaddons/modulebase.py
    """
    errmsg = ('error parsing glob expression \'%s\'' %expr +
                ' (supported glob syntax: swp1-10.300 or swp[1-10].300' +
                '  or swp[1-10]sub[0-4].300')
    regexs = ifupdown2_glob_regexs

    if regexs[0].match(expr):
        # the first regex checks for exactly two levels of ranges defined only with square brackets
        # (e.g. swpxyz[10-23]subqwe[0-4].100) to handle naming with two levels of port names.
        m = regexs[0].match(expr)
        mlist = m.groups()
        if len(mlist) < 7:
            # we have problems and should not continue
            raise Exception('%s: error: unhandled glob expression %s\n%s' % (ifacename, expr,errmsg))

        prefix = mlist[0]
        suffix = mlist[6]
        start_index = int(mlist[1])
        end_index = int(mlist[2])
        sub_string = mlist[3]
        start_sub = int(mlist[4])
        end_sub = int(mlist[5])
        for i in range(start_index, end_index + 1):
            for j in range(start_sub, end_sub + 1):
                yield prefix + '%d%s%d' % (i,sub_string,j) + suffix

    elif regexs[1].match(expr) or regexs[2].match(expr):
        # the second regex for 1 level with a range (e.g. swp10-14.100
        # the third regex checks for 1 level with [] (e.g. swp[10-14].100)
        start_index = 0
        end_index = 0
        if regexs[1].match(expr):
            m = regexs[1].match(expr)
        else:
            m = regexs[2].match(expr)
        mlist = m.groups()
        if len(mlist) != 4:
            raise Exception('%s: ' %ifacename + errmsg + '(unexpected len)')
        prefix = mlist[0]
        suffix = mlist[3]
        start_index = int(mlist[1])
        end_index = int(mlist[2])
        for i in range(start_index, end_index + 1):
            yield prefix + '%d' %i + suffix

    else:
        # Could not match anything.
        log.warning('%s: %s' %(ifacename, errmsg))
        yield expr


def ifname_is_glob(ifname):
    """
    Return True if the given name refers to multiple interfaces (a.k.a. ifname is a glob).
    Otherwise, return False.
    """
    if not ifname or ' ' in ifname:
        return False

    # "server01", for example, is not a glob.
    if re.match(r"""^\w+0\d+(?:-0\d+)?$""", ifname) is not None:
        return False

    # These are globs:
    #  swp1-5,swp8
    #  swp1-5,8
    #  swp1-5,swp7-8
    #  swp1-5,swp7
    if ',' in ifname:
        glob_parts = ifname.split(',')
        base_swp = glob_parts[0]

        re_swp = re.compile(r"""[a-zA-Z]+\d+""")
        re_number_number = re.compile(r"""\d+-\d+""")

        if ifupdown2_ifname_is_glob(base_swp):
            for part in glob_parts[1:]:
                if not (
                           part.isdigit() or
                           ifupdown2_ifname_is_glob(part) or
                           (re_swp.match(part) is not None) or
                           (re_number_number.match(part) is not None)
                       ):
                    return False

            return True

        if ifname_is_valid(base_swp):
            for part in glob_parts[1:]:
                if not (
                           part.isdigit() or
                           ifupdown2_ifname_is_glob(part) or
                           (re_swp.match(part) is not None) or
                           (re_number_number.match(part) is not None)
                       ):
                    return False

            return True

    # swp1-5
    return ifupdown2_ifname_is_glob(ifname)


def ifname_expand_glob(ifname):
    """
    Given a glob of interface names, return a list of strings containing each interface.
    """
    assert isinstance(ifname, str) or isinstance(ifname, unicode)
    ifnames = []

    # swp1-5,swp8,swp9
    # swp1-5,8,9
    # swp1-5,swp8-9
    # swp1,swp8,swp9
    # swp1,8,9
    # swp1,swp8-9
    if ',' in ifname:
        glob_parts = ifname.split(',')
        base_swp = glob_parts[0]

        if ifupdown2_ifname_is_glob(base_swp):

            # base_swp is a glob.  Expand it, and take the first interface as our new base_swp.
            modbase_expansion = []
            for name in ifupdown2_parse_glob('foobarblah', base_swp):  # ifupdown2_parse_glob returns a generator
                modbase_expansion.append(name.strip())
            base_swp = modbase_expansion[0]
            ifnames.extend(modbase_expansion)

        elif ifname_is_valid(base_swp):
            ifnames.append(base_swp)
        else:
            base_swp = None

        if base_swp:
            base_swp = base_swp.rstrip('1234567890')

            for part in glob_parts[1:]:
                re_swp = re.search(r"""^[a-zA-Z]+\d+(?:[a-zA-Z]+\d+)?$""", part)
                re_number_number = re.search(r"""(\d+)-(\d+)""", part)

                if ifupdown2_ifname_is_glob(part):
                    ifnames.extend(ifupdown2_parse_glob('foobarblah', part))

                elif part.isdigit():
                    ifnames.append(base_swp + part)

                elif re_swp:
                    ifnames.append(part)

                elif re_number_number:
                    min_x = int(re_number_number.group(1))
                    max_x = int(re_number_number.group(2)) + 1

                    for y in xrange(min_x, max_x):
                        ifnames.append(base_swp + str(y))
                else:
                    log.info("unknown glob component '{0}' in ifname_expand_glob()".format(part))

    # swp1-5
    elif ifupdown2_ifname_is_glob(ifname):
        for name in ifupdown2_parse_glob('foobarblah', ifname):
            ifnames.append(name)

    return ifnames


def ifnames_to_glob(ifnames):
    """
    Given list of strings containing interface names, return a glob.
    """
    if len(ifnames) == 1:
        return ifnames[0]

    ifnumbers = []

    # First make sure all interfaces begin with the same prefix and that none
    # of these are sub-interfaces (no . in the name)
    prev_prefix = None
    for ifname in ifnames:
        # server01....we cannot change this to server1
        re_glob_fakeout = re.match(r"""^\w+0\d+$""", ifname)
        re_glob_fakeout2 = re.match(r"""^\w+0\d+-0\d+$""", ifname)

        re_ifname = re.match(r"""^(.*?)(\d.*)$""", ifname)

        if re_glob_fakeout or re_glob_fakeout2:
            pass

        # swp1, swp2, etc
        elif re_ifname:
            prefix = re_ifname.group(1)
            suffix = re_ifname.group(2)

            # not supported for now
            if '.' in suffix:
                return None

            if not suffix.isdigit():
                return None

            if prev_prefix is not None and prefix != prev_prefix:
                return None

            ifnumbers.append(int(suffix))
            prev_prefix = prefix

    if prev_prefix and ifnumbers:
        return prev_prefix + numbers_to_glob(ifnumbers)
    return None


def numbers_to_glob(ids):
    """
    For example, given an iterable:
     (3,4,7,10,11,12,17,22,4001,4002,4003,7777,8000,8001,8002,8003,8004)

    Return a string:
     3-4,7,10-12,17,22,4001-4003,7777,8000-8004

    The parameter, ids, is an iterable that might contain integers, integers as strings, or
    integer ranges as strings.  This method handles all these types.
    """
    ids_final = []

    # Convert the input, which can contain integers and strings, to a sorted list of integers.
    for _id in ids:  # Note that "id" is a Python built-in function.
        if isinstance(_id, int):
            ids_final.append(_id)
        elif _id.isdigit():
            ids_final.append(int(_id))
        else:
            re_number_range = re.match(r"""(\d+)-(\d+)""", _id)
            if re_number_range is not None:
                start = int(re_number_range.group(1))
                end = int(re_number_range.group(2)) + 1
                if end < start:
                    print "ERROR: invalid numeric range"
                    return []
                ids_final.extend(range(start, end))

    if not ids_final:
        print "ERROR: numbers_to_glob() could not extract any IDs from %s" % pformat(ids)
        return ''

    ids = sorted(set(ids_final))  # A list
    min_id = ids[0]
    max_id = ids[-1]
    limiters = []

    # If the range of integers is continuous (i.e. 1,2,3) . . .
    if len(ids) == (max_id - min_id + 1):
        if min_id == max_id:
            limiters.append(str(min_id))
        else:
            limiters.append("{0}-{1}".format(min_id, max_id))
    # Otherwise, some integers get skipped (i.e. 1,2,5).
    else:
        streak_start_x = 0
        streak_end_x = 0
        len_ids = len(ids)

        for x in range(len_ids):

            # The first number in the list
            if x == 0:
                current = ids[x]
                next_ids = ids[x + 1]

                if current + 1 == next_ids:
                    streak_start_x = current
                else:
                    limiters.append(str(current))

            # The last number in the list
            elif x == len_ids - 1:
                prev_ids = ids[x - 1]
                current = ids[x]

                if prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("{0}-{1}".format(streak_start_x, streak_end_x))
                else:
                    limiters.append(str(current))

            # Any number in the middle of the list
            else:
                prev_ids = ids[x - 1]
                current = ids[x]
                next_ids = ids[x + 1]

                # In the middle of a streak...do nothing
                if prev_ids + 1 == current and current + 1 == next_ids:
                    pass

                # The end of a streak
                elif prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("{0}-{1}".format(streak_start_x, streak_end_x))
                    streak_start_x = 0
                    streak_end_x = 0

                # The beginning of a streak
                elif current + 1 == next_ids:
                    streak_start_x = current

                # A loner
                else:
                    limiters.append(str(current))

    return ','.join(limiters)


def glob_to_numbers(glob):
    """
    For example, given a string containing single numbers and ranges:
     3-4,7,10-12,17,22,4001-4003,7777,8000-8004

    Return a sorted list of deduplicated integers:
     [3,4,7,10,11,12,17,22,4001,4002,4003,7777,8000,8001,8002,8003,8004]
    """

    try:
        # Using split(',') instead of the replacement could yield empty strings in the result.
        glob = glob.replace(',', ' ').split()  # Creates a list
    # If glob isn't a string type, calling string methods raises an AttributeError.
    except AttributeError:
        # TODO - Don't ignore the exception.  This is here because somewhere this function
        # gets passed a list if strings.  Things still break for a mixed list of integers
        # and strings.
        pass

    numbers = set()

    for x in glob:
        if x.isdigit():
            numbers.add(int(x))
        else:
            re_range = re.match(r"""^(\d+)-(\d+)$""", x)

            if re_range is not None:
                min_range = int(re_range.group(1))
                max_range = int(re_range.group(2))

                if max_range >= min_range:
                    numbers.update(range(min_range, max_range + 1))
                else:
                    print "ERROR: Globs must consist of numbers or ranges.  {0} is neither. We were given glob '{1}'".format(x, glob)
                    return []
            else:
                print "ERROR: globs must consist of numbers or ranges, {0} is neither. We were given glob '{1}'".format(x, glob)
                return []

    return sorted(numbers)  # A sorted list


def is_number_glob(glob):
    """
    Return True if the input is a numeric glob, such as 1-5,10,13-15,20.
    The input can be a string or a list of strings.
    """
    if glob is None:
        return False

    try:
        # Using split(',') instead of the replacement could yield empty strings in the result.
        glob = glob.replace(',', ' ').split()  # Creates a list
    except Exception:
        # TODO - Don't ignore the exception.  This is here because somewhere this function
        # gets passed a list if strings.  Things still break for a mixed list of integers
        # and strings.
        pass

    # All elements must be digits or ranges.
    re_glob = re.compile(r"""^\d+(?:-\d+)?$""")
    for x in glob:
        if re_glob.match(x) is None:
            return False

    return True


def get_lldp():
    """
    Return the Link Layer Discovery Protocall.
    """
    lldp = {}

    if os.path.exists('/usr/sbin/lldpctl'):
        try:
            output = check_output(['/usr/sbin/lldpctl', '-f', 'json'])
            lldp_original = json.loads(output).get('lldp')

            if lldp_original and lldp_original[0]:
                for value in lldp_original[0].get('interface'):
                    lldp[value['name']] = value['chassis'][0]['name'][0]['value']

        except Exception as e:
            log.exception(e)

        for ifname in os.listdir('/sys/class/net/'):
            bonding_slaves_filename = "/sys/class/net/%s/bonding/slaves" % ifname

            if os.path.exists(bonding_slaves_filename):

                with open(bonding_slaves_filename) as f:
                    lines = list(f)

                lldp_peers = []

                if lines:
                    firstline = lines[0]

                    if firstline:
                        for bond_slave in firstline.split():
                            if bond_slave in lldp:
                                lldp_peers.append(lldp[bond_slave])

                if lldp_peers:
                    lldp_peers = sorted(list(set(lldp_peers)))
                    lldp[ifname] = ', '.join(lldp_peers)

    return lldp


def tabulate_pending_add_del_commands():
    header = ['User', 'Timestamp', 'Command']
    data = []

    if os.path.exists(PENDING_COMMANDS_FILE):
        with open(PENDING_COMMANDS_FILE) as f:
            for line in f:
                (user, timestamp, command) = line.strip().split('\t')
                data.append((user, timestamp, command))

    return tabulate(data, header)


def print_pending_add_del_commands():
    print "Please run 'net abort' and use net to re-issue your changes."

    if os.path.exists(PENDING_COMMANDS_FILE):
        print "\nThe following is a list of all add/del"
        print "commands since the last 'net commit'."
        print "======================================"

        with open(PENDING_COMMANDS_FILE) as f:
            for line in f:
                (user, timestamp, command) = line.strip().split('\t')
                print command
        print ''


def globify_interface_commands(init_commands):
    """
    Given:
    net add int swp1 mtu 1500
    net add int swp2 mtu 1500
    net add int swp3 mtu 1500

    Return:
    net add int swp1-3 mtu 1500
    """
    final_commands = []
    globbed_commands = []
    non_globbed_commands = []
    commands_by_id = OrderedDict()

    for command in init_commands:
        command_list = command.split()

        # net add bridge
        if len(command_list) <= 3:
            final_commands.append(command)
            continue

        add_del = command_list[1]
        iface_type = command_list[2]

        # net add vlan 10-20
        if iface_type == 'vlan' and len(command_list) == 4:
            final_commands.append(command)
            continue

        ifname = command_list[3]
        command_id = (add_del, iface_type, ' '.join(command_list[4:]))

        if command_id not in commands_by_id:
            commands_by_id[command_id] = []
        commands_by_id[command_id].append(ifname)

    # We must first configure the bond-slaves, vlan-interface, and other statements.
    for bond_slaves_etc_first in (True, False):
        for (command_id, ifnames) in commands_by_id.iteritems():
            (add_del, iface_type, keyword_value) = command_id

            if (keyword_value.startswith('bond-slaves') or
                keyword_value.startswith('vxlan')):
                if not bond_slaves_etc_first:
                    continue
            else:
                if bond_slaves_etc_first:
                    continue

            if len(ifnames) > 1:
                ifnames_glob = ifnames_to_glob(ifnames)
            else:
                ifnames_glob = None

            if ifnames_glob:
                globbed_commands.append("net {0} {1} {2} {3}".format(add_del, iface_type, ifnames_glob, keyword_value))
            else:
                for ifname in ifnames:
                    non_globbed_commands.append("net {0} {1} {2} {3}".format(add_del, iface_type, ifname, keyword_value))

    final_commands.extend(globbed_commands)
    final_commands.extend(non_globbed_commands)
    return final_commands


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def seconds_to_hhmmss(secs):
    # TODO - Can I delete code by using the Python standard library "time" package?
    if not secs:
        return 'never'

    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    days, hours = divmod(hours, 24)

    if days:
        return '{0} day{1}, {2:0>2}:{3:0>2}:{4:0>2}'.format(days, 's' if days > 1 else '',  hours, mins, secs)
    else:
        return '{0:0>2}:{1:0>2}:{2:0>2}'.format(hours, mins, secs)


def get_hostname(ip):
    try:
        hostname = gethostbyaddr(ip)
        return " ({0})".format(hostname[0])
    except Exception:
        return ""


def file_set_key_value(filename, keyword, new_value, comment_out=False):
    """
    For config files that follow a key=value notation (i.e. ip = 10.1.1.1), this function
    provides a way to alter the value for a key.
    """

    if not os.path.exists(filename):
        log.info("Could not set {0} to {1} in {2}.  The file does not exist.".format(keyword, new_value, filename))
        return

    new_content = []

    with open(filename) as f:
        for line in f:
            # This regex matches the case when a key is present but not assigned a value (i.e. key=<blank> # Comment).
            re_key_no_value = re.search(r'''^(?:\s*#)?(\s*{0}\s*)=[\t ]*(#[\S\t ]*)?$'''.format(keyword), line, re.IGNORECASE)

            if re_key_no_value:
                before_equals = re_key_no_value.group(1)
                comment = re_key_no_value.group(2)
                line = ('#{0}= {1}' if comment_out else '{0}= {1}').format(before_equals, new_value)
                if comment:
                    line += ' ' + comment
                line += '\n'
            else:
                # This regex matches the case when a key is present and assigned a value (i.e. key=value # Comment).
                re_key_value = re.search(r'''^(?:\s*#)?(\s*{0}\s*=\s*)(?:".*"|\S[^#\s]*)([\t ]*#[\S\t ]*)?$'''.format(keyword), line, re.IGNORECASE)

                if re_key_value:
                    before_value = re_key_value.group(1)
                    comment = re_key_value.group(2)
                    line = ('#{0}{1}' if comment_out else '{0}{1}').format(before_value, new_value)
                    if comment:
                        line += comment
                    line += '\n'

            new_content.append(line)

    with open(filename, 'w') as f:
        f.write(''.join(new_content))


def is_mac_address(word):
    return (re.match('(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}', word) or
            re.match('[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}', word))


class ConfigFile(object):

    def __init__(self, parent, filename, cmds):
        """
        'parent' is the ConfigWrapper class that owns us.

        TODO - Passing the parent class is not necessary.
        http://stackoverflow.com/questions/10091957/get-parent-class-name#10092179

        'cmds' is a list of commands to run if the configuration file is modified.
        """

        self.parent = parent
        self.filename = filename.strip()

        filename_underscore = self.filename.replace('/', '_').replace('.', '_').strip('_')
        self.scratchpad = os.path.join(self.parent.var_run_nclu, "%s_scratchpad" % filename_underscore)
        self.backup = os.path.join(self.parent.var_run_nclu, "%s_backup" % filename_underscore)
        self.cmds = cmds

    def __str__(self):
        return self.filename

    def remove_backup(self):
        if os.path.exists(self.backup):
            os.remove(self.backup)

    def remove_scratchpad(self):
        if os.path.exists(self.scratchpad):
            os.remove(self.scratchpad)


class ConfigFileSingleLine(ConfigFile):
    """
    For config files with a single line
    """

    def edit(self, value):
        """
        Return True if the file was modified
        """
        current_value = self.get_value(True)
        if current_value == value:
            print "{0} already has {1}".format(self.scratchpad, value)
            return False
        else:
            with open(self.scratchpad, 'w') as f:
                f.write(value + '\n')
            return True

    def get_value(self, read_scratchpad=False):
        """
        This file should only contain one line. If it happens to contain more
        than one return the first valid (not blank and not a comment) and log
        a warning.
        """
        result = None
        filename = self.scratchpad if read_scratchpad else self.filename

        if os.path.exists(filename):
            with open(filename) as f:
                for line in f.readlines():
                    line = line.strip()

                    if line and not line.startswith('#'):
                        if result is None:
                            result = line
                        else:
                            log.warning("{0} contains multiple values.  It should only contain one value.".format(filename))

        return result


class ConfigFileKeyValue(ConfigFile):

    def edit(self, key, value, comment_out=False):
        """
        Return True if the file was modified.
        """
        current_value = self.get_value(True)  # A dictionary

        if current_value.get(key) == value:
            print "{0} already has '{1} = {2}'".format(self.scratchpad, key, value)
            return False
        else:
            file_set_key_value(self.scratchpad, key, value, comment_out)
            return True

    def get_value(self, read_scratchpad=False):
        filename = self.scratchpad if read_scratchpad else self.filename
        results = {}

        if os.path.exists(filename):
            with open(filename) as f:
                for line in f:
                    re_key_value = re.search(r'''^\s*(\w+)\s*=\s*"?\s*(.*?)\s*"?\s*$''', line, re.IGNORECASE)

                    if re_key_value:
                        key = re_key_value.group(1)
                        value = re_key_value.group(2)
                        results[key] = value
        return results


class ConfigWrapperBase(object):

    def __init__(self, name, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs
        self.name = name
        self.var_run_nclu = "/var/run/nclu/{0}".format(self.name)

        if not os.path.exists(self.var_run_nclu):
            mkdir_p(self.var_run_nclu)

        self.files = {}  # The keys are file names; the values are ConfigFile instances (including subclasses).

    def add_config_file(self, filename, cmd):
        self.files[filename] = ConfigFile(self, filename, cmd)

    def get_managed_files(self):
        return [x.filename for x in self.files.itervalues()]

    def get_service_cmd_for_file(self, filename):
        cfg = self.files.get(filename)
        return cfg.cmds if cfg else []

    def abort_pending(self):
        if os.path.exists(self.var_run_nclu):
            for filename in os.listdir(self.var_run_nclu):
                full_filename = os.path.join(self.var_run_nclu, filename)

                if os.path.isdir(full_filename):
                    shutil.rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def del_all(self, cli):
        raise NotImplementedError("Subclasses should override the del_all method.")

    def commit_pending(self, verbose):
        files_modified = []
        commit_ok = True
        cmds = []

        # Copy scratchpad files over the real config files
        for cfg in self.files.itervalues():
            if os.path.exists(cfg.scratchpad) and not files_match(cfg.filename, cfg.scratchpad):
                shutil.copy(cfg.filename, cfg.backup)
                shutil.copy(cfg.scratchpad, cfg.filename)
                files_modified.append(cfg)
                cmds.extend(cfg.cmds)

        # Run all commands that are needed to apply the changes in the new config files
        for (cmd_index, cmd) in enumerate(cmds):
            try:
                output = check_output(cmd, stderr=STDOUT)

                if verbose:
                    print output

            except Exception as e:
                print "\nERROR: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                commit_ok = False
                break

        if commit_ok:
            # Remove scratchpad and backup files
            for cfg in files_modified:
                cfg.remove_scratchpad()
                cfg.remove_backup()

            return (True, [cfg.filename for cfg in files_modified])

        else:
            # Restore the config file from our backup
            for cfg in files_modified:
                shutil.move(cfg.backup, cfg.filename)

            # Rerun the commands we ran to apply config changes
            for cmd in cmds[cmd_index]:
                try:
                    output = check_output(cmd, stderr=STDOUT)

                    if verbose:
                        print output

                except Exception as e:
                    print "\nERROR: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)

            return (False, [])

    def eval_command_line_args(self, cli):
        # If we are doing an add or del, we must create the scratchpad file.
        if cli.args.get('add') or cli.args.get('del'):
            for cfg in self.files.itervalues():
                if not os.path.exists(cfg.scratchpad):
                    shutil.copyfile(cfg.filename,  cfg.scratchpad)

        return True
        # The child class will need to do additional work in its eval_command_line_args().
        # The overrided function should return True to inform the calling block that the
        # command was successfully executed.  Among other consequences of not returning True,
        # the command will not show up among the pending commands in the output of "net pending".

    def show_pending(self, print_errors=True):
        diff_pending = False

        for cfg in self.files.itervalues():
            if os.path.exists(cfg.scratchpad):
                if print_diff(cfg.filename, cfg.scratchpad, self.color_diffs):
                    diff_pending = True

        return diff_pending

    def show_config_files(self, user_may_edit):

        for cfg in self.files.itervalues():
            if os.path.exists(cfg.filename) and os.path.getsize(cfg.filename) > 0:
                print cfg.filename
                print '=' * len(cfg.filename)
                print_file(cfg.filename)

    def show_config_commands(self, user_may_edit):
        raise NotImplementedError("Subclasses should override the show_config_commands method.")

    def show_config_summary(self, user_may_edit, summary):
        raise NotImplementedError("Subclasses should override the show_config_summary method.")


def get_interfaces():
    """
    Get a full list by combining the interfaces in /etc/network/interfaces
    along with any that are pending.
    """
    ifnames = set()

    if os.path.exists('/var/run/nclu/iface/'):
        for filename in os.listdir('/var/run/nclu/iface/'):
            if not filename.endswith('.original') and not filename.endswith('.del'):
                ifnames.add(filename)

    if os.path.exists('/etc/network/interfaces'):
        with open('/etc/network/interfaces') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith('iface '):
                    ifname = line.split()[1]
                    ifnames.add(ifname)

    if os.path.exists('/sys/class/net/'):
        for ifname in os.listdir('/sys/class/net/'):
            if ifname not in ('swid0_eth', 'bonding_masters') and not ifname.endswith('-v0'):
                ifnames.add(ifname)

    return sort_for_humans(list(ifnames))


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = decode_dict(item)
        rv.append(item)
    return rv


def decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = decode_dict(value)
        rv[key] = value
    return rv
